/// An finite orderered collection of characters.
abstract class Alphabet {
  factory Alphabet.fromSet(Set<String> fromSet) =>
      new _DefaultAlphabet(fromSet);

  /// The first symbol in this alphabet.
  String get start;

  /// The last symbol in this alphabet.
  String get end;

  /// Returns true iff [char] is in this alphabet.
  bool contains(String char);

  /// Returns the successor to [char] in this alphabet or null if [char] == end
  /// or [char] is not in this alphabet.
  String successor(String char);

  /// Returns the predecessor to [char] in this alphabet or null if [char] ==
  /// start or [char] is not in this alphabet.
  String predecessor(String char);
}

/// A delegating [Alphabet] implementation.
class AlphabetTransformer implements Alphabet {
  final Alphabet _delegate;

  AlphabetTransformer(this._delegate);

  @override
  String get start => _delegate.start;

  @override
  String get end => _delegate.end;

  @override
  String successor(String char) => _delegate.successor(char);

  @override
  String predecessor(String char) => _delegate.predecessor(char);

  @override
  bool contains(String char) => _delegate.contains(char);
}

/// An [Alphabet] that iterates over a specified set of characters.
class _DefaultAlphabet implements Alphabet {
  final List<String> _symbols;

  _DefaultAlphabet(Set<String> symbols) : this._symbols = symbols.toList();

  @override
  bool contains(String char) => _symbols.contains(char);

  @override
  String get end => _symbols.last;

  @override
  String get start => _symbols.first;

  @override
  String predecessor(String char) {
    if (_symbols.contains(char)) {
      int index = _symbols.indexOf(char);
      return index > 0 ? _symbols[index - 1] : null;
    }
    return null;
  }

  @override
  String successor(String char) {
    if (_symbols.contains(char)) {
      int index = _symbols.indexOf(char);
      return index < _symbols.length - 1 ? _symbols[index + 1] : null;
    }
    return null;
  }
}
