import 'package:string_sequence/common_alphabets.dart';
import 'package:string_sequence/string_sequence.dart';

/// A [StringSequence] that deterministically iterates over an infinite sequence
/// of strings which are created by looping through an alphabet.
///
/// For example, if the alphabet consisted of all lowercase letters in the
/// English alphabet with the first string being 'a', execution would look like
/// this.
///
///     nextString() // 'a'
///     nextString() // 'b'
///     ...
///     nextString() // 'z'
///     nextString() // 'aa'
///     nextString() // 'ab'
///     ...
///     nextString() // 'az'
///     ...
///     nextString() // 'ba'
///     ...
///     nextString() // 'zz'
///     nextString() // 'aaa'
///     nextString() // 'aab'
class EnumeratingStringSequence implements StringSequence {
  final LoopingAlphabet _alphabet;
  String _string = '';

  EnumeratingStringSequence(this._alphabet);

  /// Creates an [EnumeratingStringSequence] which starts iteration after
  /// [initialString].
  ///
  /// It is assumed that [initialString] is a valid string in this sequence.
  EnumeratingStringSequence.startingAfter(String initialString, this._alphabet)
      : _string = initialString;

  /// Generates the next string in the sequence.
  @override
  String next() {
    if (_string.isEmpty) {
      _string = _alphabet.start;
      return _string;
    }

    // If the string consists of only _alphabet.end characters, increase the
    // length of the string with _alphabet.end before we increment characters.
    if (_string.split('').every((s) => s == _alphabet.end)) {
      _string = '$_string${_alphabet.end}';
    }

    // Walk backward through the string incrementing characters. If a character
    // loops to the start of the alphabet, increment its left neighbor.
    var position = _string.length - 1;
    bool shouldIncrementNeighbor;
    do {
      var nextChar = _alphabet.successor(_string[position]);
      _string = _string.replaceRange(position, position + 1, nextChar);
      shouldIncrementNeighbor = --position >= 0 && nextChar == _alphabet.start;
    } while (shouldIncrementNeighbor);
    return _string;
  }
}
